/**
 * ncp.c, creates strange textures 
 *  Version 0.2
 * Copyright (c) 1999 by Stefan Stiasny <sc@gimp.org>
 * 
 * thanx to sid for inspiration
 * gtk code derived from snoise.c 
 *             Copyright (C) 1997, 1998 Marcelo de Gomensoro Malheiros
 *
 * The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 **/



#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <limits.h>

#include "gtk/gtk.h"
#include "libgimp/gimp.h"


#ifndef RAND_MAX
#define RAND_MAX 2147483647
#endif /* RAND_MAX */

#define ENTRY_WIDTH 40

typedef struct {
  gint    seed;
  gint    timeseed;
  gint    points;
  gint    closest;
  gint    color;
} ncpValues;

typedef struct {
   gint run;
} ncpInterface;

typedef struct {
  int x, y;
} point;

/*---- Prototypes ----*/

static void query (void);
static void run (char *name, int nparams, GParam  *param, int *nreturn_vals, GParam **return_vals);

static void ncp (GDrawable *drawable);

static gint ncp_dialog (void);
static void dialog_close_callback (GtkWidget *widget, gpointer data);
static void dialog_toggle_update (GtkWidget *widget, gpointer data);
static void dialog_entry_callback (GtkWidget *widget, gpointer data);
static void dialog_ok_callback (GtkWidget *widget, gpointer data);


/*---- Variables ----*/

GPlugInInfo PLUG_IN_INFO =
{
  NULL,  /* init_proc */
  NULL,  /* quit_proc */
  query, /* query_proc */
  run,   /* run_proc */
};

static ncpValues ncpvals = {
  1,   /* seed */
  0,   /* use time seed */
  12,  /* points */
  1,   /* nth closest point */
  0    /* use color */
};

static ncpInterface ncpint = {
  FALSE /* run */
};


/* FUNCTIONS */

MAIN()

static void
query (void)
{
  static GParamDef args[] =
  {
    { PARAM_INT32, "run_mode", "Interactive, non-interactive" },
    { PARAM_IMAGE, "image", "Input image" },
    { PARAM_DRAWABLE, "drawable", "Input drawable" },
    { PARAM_INT32, "seed", "Random seed" },
    { PARAM_INT32, "points", "Number of points" },
  };
  static GParamDef *return_vals = NULL;
  static gint nargs = sizeof (args) / sizeof (args[0]);
  static gint nreturn_vals = 0;

  gimp_install_procedure ("plug_in_ncp",
           "Creates a grayscale texture",
           "Generates 2D textures",
           "",
           "",
           "Apr 1998, v1.03",
           "<Image>/Filters/Render/ncp",
           "RGB*, GRAY*",
           PROC_PLUG_IN,
           nargs, nreturn_vals, args, return_vals);
}

static void
run (char *name, int nparams, GParam *param, int *nreturn_vals,
     GParam **return_vals) {
   static GParam values[1];
  
     GDrawable *drawable;
     GRunModeType run_mode;
     GStatusType status;
  
     status = STATUS_SUCCESS;
     run_mode = param[0].data.d_int32;

     values[0].type = PARAM_STATUS;
     values[0].data.d_status = status;

     *nreturn_vals = 1;
     *return_vals  = values;
  
     /*  Get the specified drawable  */
     drawable = gimp_drawable_get (param[2].data.d_drawable);

     /*  See how we will run  */
     switch (run_mode) {
     case RUN_INTERACTIVE:
       /*  Possibly retrieve data  */
       gimp_get_data("plug_in_ncp", &ncpvals);

       /*  Get information from the dialog  */
       if (!ncp_dialog ()) return;
       break;

   case RUN_NONINTERACTIVE:
      /*  Test number of arguments  */

// FIX THIS !!!!

      if (nparams == 9) {
      ncpvals.seed = param[3].data.d_int32;
      ncpvals.timeseed = param[4].data.d_int32;
      ncpvals.points = param[5].data.d_int32;
      ncpvals.closest = param[6].data.d_int32;
      ncpvals.color = param[7].data.d_int32;
    }
    else
      status = STATUS_CALLING_ERROR;
    break;

  case RUN_WITH_LAST_VALS:
    /*  Possibly retrieve data  */
    gimp_get_data("plug_in_ncp", &ncpvals);
    break;
    
  default:
    break;
  }
  
  /*  Create texture  */
  if ((status == STATUS_SUCCESS) && (gimp_drawable_color (drawable->id) ||
      gimp_drawable_gray (drawable->id)))
    {
      /*  Set the tile cache size  */
      gimp_tile_cache_ntiles((drawable->width + gimp_tile_width() - 1) / gimp_tile_width());

      /*  Run!  */
      ncp (drawable);

      /*  If run mode is interactive, flush displays  */
      if (run_mode != RUN_NONINTERACTIVE)
        gimp_displays_flush();

      /*  Store data  */
      if (run_mode==RUN_INTERACTIVE || run_mode==RUN_WITH_LAST_VALS)
        gimp_set_data("plug_in_ncp", &ncpvals,
                      sizeof(ncpValues));
    }
  else
    {
      /* gimp_message ("ncp: cannot operate on indexed color images"); */
      status = STATUS_EXECUTION_ERROR;
    }  
  
  values[0].data.d_status = status;
  
  gimp_drawable_detach(drawable);
}



static gint
ncp_dialog (void)
{
  GtkWidget *dlg;
  GtkWidget *table;
  GtkWidget *button;
  GtkWidget *label;
  GtkWidget *entry;
  GtkWidget *seed_hbox;
  GtkWidget *time_button;
  GtkWidget *color_cb;
  gchar **argv;
  gint  argc;
  gchar buffer[32];

  /*  Set args  */
  argc = 1;
  argv = g_new (gchar *, 1);
  argv[0] = g_strdup ("ncp");
  gtk_init (&argc, &argv);
  gtk_rc_parse (gimp_gtkrc ());

  /*  Dialog initialization  */
  dlg = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (dlg), "ncp");
  gtk_window_position (GTK_WINDOW (dlg), GTK_WIN_POS_MOUSE);
  gtk_container_border_width (GTK_CONTAINER (GTK_DIALOG (dlg)->action_area), 2);
  gtk_signal_connect (GTK_OBJECT (dlg), "destroy",
                      (GtkSignalFunc) dialog_close_callback, NULL);

  /*  Table  */
  table = gtk_table_new (4, 3, FALSE);
  gtk_container_border_width (GTK_CONTAINER (table), 2);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->vbox), table, TRUE, TRUE, 0);

  /* SEED LABEL */
  label = gtk_label_new ("Seed");
  gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 0, 1,
                    GTK_FILL, GTK_FILL, 1, 0);
  gtk_widget_show (label);
  
  seed_hbox = gtk_hbox_new (FALSE, 2);
  gtk_table_attach (GTK_TABLE (table), seed_hbox, 1, 2, 0, 1,
		    GTK_FILL, GTK_FILL, 0, 0);

  /* SEED ENTRY */
  entry = gtk_entry_new ();
  gtk_box_pack_start (GTK_BOX (seed_hbox), entry, TRUE, TRUE, 0);
  gtk_widget_set_usize (entry, ENTRY_WIDTH, 0);
  sprintf(buffer, "%d", ncpvals.seed);
  gtk_entry_set_text (GTK_ENTRY (entry), buffer);
  gtk_signal_connect (GTK_OBJECT (entry), "changed",
                      (GtkSignalFunc) dialog_entry_callback, &ncpvals.seed);
  gtk_widget_show (entry);

  /* TIME SEED BUTTON */
  time_button = gtk_toggle_button_new_with_label ("Time");
  gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON(time_button), ncpvals.timeseed);
  gtk_signal_connect (GTK_OBJECT (time_button), "toggled",
		      (GtkSignalFunc) dialog_toggle_update,
		      &ncpvals.timeseed);
  gtk_box_pack_end (GTK_BOX (seed_hbox), time_button, FALSE, FALSE, 0);
  gtk_widget_show (time_button);
  gtk_widget_show (seed_hbox);


  /* POINTS  */
  label = gtk_label_new ("Points");
  gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 1, 2,
                    GTK_FILL, GTK_FILL, 1, 0);
  gtk_widget_show (label);
  
  entry = gtk_entry_new ();
  gtk_table_attach (GTK_TABLE (table), entry, 1, 2, 1, 2,
                    GTK_FILL, GTK_FILL, 0, 0);
  gtk_widget_set_usize (entry, ENTRY_WIDTH, 0);
  sprintf(buffer, "%d", ncpvals.points);
  gtk_entry_set_text (GTK_ENTRY (entry), buffer);
  gtk_signal_connect (GTK_OBJECT (entry), "changed",
                      (GtkSignalFunc) dialog_entry_callback, &ncpvals.points);
  gtk_widget_show (entry);

 
  /* CLOSEST  */
  label = gtk_label_new ("foo-point ");
  gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, 2, 3,
                    GTK_FILL, GTK_FILL, 1, 0);
  gtk_widget_show (label);
  
  entry = gtk_entry_new ();
  gtk_table_attach (GTK_TABLE (table), entry, 1, 2, 2, 3,
                    GTK_FILL, GTK_FILL, 0, 0);
  gtk_widget_set_usize (entry, ENTRY_WIDTH, 0);
  sprintf(buffer, "%d", ncpvals.closest);
  gtk_entry_set_text (GTK_ENTRY (entry), buffer);
  gtk_signal_connect (GTK_OBJECT (entry), "changed",
                      (GtkSignalFunc) dialog_entry_callback, &ncpvals.closest);
  gtk_widget_show (entry); 


  /* COLOR CHECK */
  color_cb = gtk_check_button_new_with_label ("Color");
//  gtk_misc_set_alignment (GTK_MISC (color_cb), 0.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), color_cb, 0, 3, 3, 4,
                    GTK_FILL, GTK_FILL, 1, 0);
 gtk_toggle_button_set_state (GTK_TOGGLE_BUTTON(color_cb), ncpvals.color);
  gtk_signal_connect (GTK_OBJECT (color_cb), "toggled",
		      (GtkSignalFunc) dialog_toggle_update,
		      &ncpvals.color);
  gtk_widget_show (color_cb);


  /*  OK BUTTON  */
  button = gtk_button_new_with_label ("OK");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect (GTK_OBJECT (button), "clicked",
                      (GtkSignalFunc) dialog_ok_callback, dlg);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->action_area), button,
                      TRUE, TRUE, 0);
  gtk_widget_grab_default (button);
  gtk_widget_show (button);

  /*  CANCEL BUTTON   */
  button = gtk_button_new_with_label ("Cancel");
  GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) gtk_widget_destroy,
                             GTK_OBJECT (dlg));
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->action_area), button,
                      TRUE, TRUE, 0);
  gtk_widget_show (button);

  gtk_widget_show (table);
  gtk_widget_show (dlg);

  gtk_main ();
  gdk_flush ();

  return ncpint.run;
}


static void
dialog_close_callback (GtkWidget *widget, gpointer data)
{
  gtk_main_quit ();
}


static void
dialog_toggle_update (GtkWidget *widget, gpointer data)
{
  int *toggle_val;

  toggle_val = (int *) data;

  if (GTK_TOGGLE_BUTTON (widget)->active)
    *toggle_val = TRUE;
  else
    *toggle_val = FALSE;
}


static void
dialog_entry_callback (GtkWidget *widget, gpointer data)
{
  gint *text_val;

  text_val = (gint *) data;

  *text_val = atoi (gtk_entry_get_text (GTK_ENTRY (widget)));
}



static void
dialog_ok_callback (GtkWidget *widget, gpointer data)
{
  ncpint.run = TRUE;
  gtk_widget_destroy (GTK_WIDGET (data));
}

int compare(const void *a, const void *b) {

return (*(int *)a - *(int *)b);

/*  if ((*(int *)a)<(*(int *)b))  return -1;
  if ((*(int *)a)==(*(int *)b))  return -1;
  return 1;
*/
}

static void
ncp(GDrawable *drawable){ // int width, int height, int seed) {

gint i, k;
gint xmid, ymid;

point *p, *vp; 

GPixelRgn dest_rgn;
gint color, chns, channels, l, has_alpha, row, col;
gint sel_x1, sel_y1, sel_x2, sel_y2;
gint sel_width, sel_height;
gint width, height;
gint progress, max_progress;
gpointer pr;
guchar *dest, *dest_row;
guchar val;
gint points;
int *distances;

  points = ncpvals.points;
  color = ncpvals.color;

  /*  Get selection area  */
  gimp_drawable_mask_bounds (drawable->id, &sel_x1, &sel_y1, &sel_x2, &sel_y2);

/*
  fprintf(stderr, "drawable: %d von ( x:%d / y:%d ) bis ( x:%d / y:%d ) \n", drawable->id, sel_x1, sel_y1, sel_x2, sel_y2); 
*/

  sel_width = sel_x2 - sel_x1;
  sel_height = sel_y2 - sel_y1;

  fprintf(stderr, "points:%d w:%d h:%d \n", points, sel_width, sel_height);

  /*  Force sane parameters  */
  if (ncpvals.points < 1) ncpvals.points = 1;
  if (ncpvals.points > 200) ncpvals.points = 200;
  if (ncpvals.seed < 0) ncpvals.seed = 0;
  if (ncpvals.closest < 1) ncpvals.closest = 1;
  if (ncpvals.closest > ncpvals.points) ncpvals.closest = ncpvals.points;
  
/*
  fprintf(stderr, "points: %d seed :%d timeseed: %d color: %d closest: %d\n",
   ncpvals.points, ncpvals.seed, ncpvals.timeseed, ncpvals.color,ncpvals.closest);
*/

  /*  Define the pseudo-random number generator seed  */
  if (ncpvals.timeseed) ncpvals.seed = time(NULL);
  srand (ncpvals.seed); 

  chns = gimp_drawable_bpp (drawable->id);
  has_alpha = 0;
  if (gimp_drawable_has_alpha (drawable->id)) {
    chns--;
    has_alpha = 1;
  }

  gimp_progress_init ("ncpifying ... ");
  progress = 0;
  max_progress = sel_width * sel_height;
 
  if (ncpvals.color!=0) max_progress *= chns;

  width=sel_width; 
  height=sel_height;

  /* allocate memory */
  distances = malloc(4*points*sizeof(int));
  p = malloc(points*sizeof( point));
  vp  = malloc(4*points*sizeof(point));

 /* repeat for each channel */
 for (channels=0; channels<chns; channels++) {
   fprintf(stdout, "Check %d\n", channels+1);

  /* generate n random points */
  for (i=0;i<points;i++) {
     p[i].x= (int) (1.0 * width * rand()/(RAND_MAX+1.0));
     p[i].y= (int) (1.0 * height * rand()/(RAND_MAX+1.0));
  }
 
  xmid = width/2;
  ymid = height/2;

  /* transform points */
  for (i=0;i<points;i++) {
    vp[i].x = p[i].x + xmid;
    vp[i].y = p[i].y + ymid;
  }

  /* create pseudo points */
  for (i=0;i<points;i++) {
   vp[i+points].x = (p[i].x < xmid) ? (vp[i].x + width) : (vp[i].x - width);
   vp[i+points].y = vp[i].y;
   vp[i+2*points].x = vp[i].x;
   vp[i+2*points].y = (p[i].y < ymid) ? (vp[i].y + height) : (vp[i].y - height);
   vp[i+3*points].x = (p[i].x < xmid) ? (vp[i].x + width) : (vp[i].x - width);
   vp[i+3*points].y = (p[i].y < ymid) ? (vp[i].y + height) : (vp[i].y - height);
  }

  /* get the pixel regions */
  gimp_pixel_rgn_init(&dest_rgn, drawable, sel_x1, sel_y1, sel_width, sel_height, TRUE, TRUE);

  /* process the pixel region tiles */
  for (pr = gimp_pixel_rgns_register(1, &dest_rgn); pr != NULL; pr=gimp_pixel_rgns_process (pr)) {
    dest_row = dest_rgn.data;
    for (row = dest_rgn.y; row < (dest_rgn.y + dest_rgn.h); row++) {
      dest = dest_row;
      for (col = dest_rgn.x; col < (dest_rgn.x + dest_rgn.w); col++) {

#if 0
 for (k=0;k<points*4;k++) {
         if (vp[k].x < col+xmid ) 
         distances[k] = (int) sqrt( (col+xmid-vp[k].x) * (col+xmid-vp[k].x) 
                                 + (row+ymid-vp[k].y) * (row+ymid-vp[k].y) );
         else distances[k] = 1000;          
      }
#endif
#if 0
      for (k=0;k<points*4;k++) {
         if (vp[k].x < col+xmid && vp[k].y < row+ymid ) 
         distances[k] = (int) sqrt( (col+xmid-vp[k].x) * (col+xmid-vp[k].x) 
                                 + (row+ymid-vp[k].y) * (row+ymid-vp[k].y) );
         else distances[k] = 1000;          
      }
#endif
#if 1
      /* compute distance to each point */
      for (k=0;k<points*4;k++) {
        distances[k] = (int) sqrt( (col+xmid-vp[k].x) * (col+xmid-vp[k].x) 
                                 + (row+ymid-vp[k].y) * (row+ymid-vp[k].y) );
      }

      /* sorting the distances */
      qsort(distances, points*4, sizeof(int), compare);
      val = distances[ncpvals.closest - 1];

#endif
#if 0
   /* compute distance to each point */
      for (k=0;k<points*4;k++) {
        distances[k] = (int) sqrt( (col+xmid-vp[k].x) * (col+xmid-vp[k].x)
                                 + (row+ymid-vp[k].y) * (row+ymid-vp[k].y) );
      }

      /* sorting the distances */
      qsort(distances, points*4, sizeof(int), compare);

      /* assing value */
      if (abs(distances[ncpvals.closest - 1] - distances[ncpvals.closest]) < 10)
         val = 0;
      else val = distances[ncpvals.closest - 1];
#endif

      /* stretching contrast */
      val = (guchar)(255.0 *  (val / sqrt(width*height))); 

      /* invert */ 
      val = (guchar) (255.0 - val);
          if (color==0) { 
            for (l=0; l<chns; l++) *dest++ = val;
            if (has_alpha) *dest++ = 255;
          }
          else {
            dest+=channels;
            *dest = val;
            dest+= chns-channels;  // skip other channels 
            if (has_alpha) *dest++ = 255;
          }
        }
        dest_row += dest_rgn.rowstride;
      }
      /*  Update progress  */
      progress += dest_rgn.w * dest_rgn.h;
      gimp_progress_update ((double) progress / (double) max_progress);
    }
  /* leave for loop if greyscale output requested */
  if (ncpvals.color == 0) break;

  } 


  gimp_drawable_flush (drawable);
  gimp_drawable_merge_shadow (drawable->id, TRUE);
  gimp_drawable_update (drawable->id, sel_x1, sel_y1, sel_width, sel_height);

  free(distances);
  free(vp);
  free(p);
}


