#!/bin/sh
# -*-Mode: Tcl;-*-
#\
exec wish "$0" "$@"

set verbose 0
set Gimptcllib Gimptcl.so

set path {
    $env(HOME)/.gimp/plug-ins
    /usr/X11R6/libexec/gimp
    /usr/X11R6/lib/gimp
    /usr/local/libexec/gimp
    /usr/local/lib/gimp
    $env(HOME)
}

set gtcl_loaded "notyet"

# find and load the Gimptcl.so share library
foreach dir $path {
    if {$verbose} {puts -nonewline "looking in $dir"}
    if {![catch "glob $dir/$Gimptcllib" fqp]} {
    if {$verbose} {puts " found."}
        set gtcl_loaded "ok"
        load [lindex $fqp 0]
        break
    }
    if {$verbose} {puts "."}
}

if {[string compare $gtcl_loaded "notyet"] == 0} {
    return -code error "couldn't find $Gimptcllib"
}

proc gimptcl_query {} {
    gimp-install-procedure "spaced_out_text" \
	"Print out some text with variable spacing.  The current foreground and\
	background colors are used to create a new layer. \
	If a negative value is \
	used for spacing, then the text is jumbled together." \
	"" \
	"Zachariah Baum <zack@studioarchetype.com>" \
	"" \
	"1999" \
        "<Image>/Filters/Render/Spaced Out Text" \
	"RGB" \
	plugin \
	{
	    {int32 "run_mode" "Interactive, Non-interactive"}
	    {image "image" "The image to build the background for"}
	    {drawable "drawable" "Unused (a new layer is created)"}
	} \
	{}
#    puts stderr "Testicle installed"
}

proc gimptcl_run {mode img drw} {
#    puts stderr "mode $mode; image: $img; drawable: $drw"
    switch $mode {
	0 {
	    global image
	    set image $img
	    build-ui
	}
	1 {
	}
	2 {
	}
    }
}

# The input form for info about the desired text
proc build-ui {} {
    frame .ctl

    frame .ctl.text -relief ridge -borderwidth 3
    label .ctl.text.label -text "Text"
    entry .ctl.text.entry -width 32
    pack .ctl.text.label .ctl.text.entry \
	-side left -fill both

    frame .ctl.font -relief ridge -borderwidth 3
    label .ctl.font.label -text "Font"
    entry .ctl.font.entry -width 32
    pack .ctl.font.label .ctl.font.entry -side left -fill both

    frame .ctl.pixel_size -relief ridge -borderwidth 3
    label .ctl.pixel_size.label -text "Pixel Size"
    entry .ctl.pixel_size.entry -width 32
    pack .ctl.pixel_size.label .ctl.pixel_size.entry -side left -fill both

    frame .ctl.spacing -relief ridge -borderwidth 3
    label .ctl.spacing.label -text "Spacing (in pixels)"
    entry .ctl.spacing.entry -width 32
    pack .ctl.spacing.label .ctl.spacing.entry -side left -fill both

    frame .ctl.transparent -relief ridge -borderwidth 3
    label .ctl.transparent.label -text "Transparent Background"
    checkbutton .ctl.transparent.checkbutton -variable transparent
    pack .ctl.transparent.label .ctl.transparent.checkbutton -side left \
	    -fill both

    pack .ctl.text .ctl.font .ctl.pixel_size .ctl.spacing .ctl.transparent \
	    -side top -fill both

    frame .ctl.cq
    # clicking the "Create" button calls the "sf" procedure which actually
    # builds the textual layer.
    button .ctl.cq.create -text Create -command {
        .ctl.cq.progress configure -text "One moment..."
	sf [.ctl.text.entry get] \
		[.ctl.font.entry get] [.ctl.pixel_size.entry get] \
		[.ctl.spacing.entry get]
        .ctl.cq.progress configure -text "Done!"
    }
    button .ctl.cq.quit -text Quit -command "destroy ."
    label .ctl.cq.progress -text "" -width 32
    pack .ctl.cq.create .ctl.cq.quit .ctl.cq.progress -side left

    pack .ctl.cq -side bottom

    pack .ctl -side left -fill both
}

proc sf {text font pixel_size spacing} {
    global transparent
    global image drawable
    if {[string length $text] == 0} {
	puts stderr "Please enter some text, damnit!"
	return
    }
    if [catch {set extents [gimp-text-get-extents $text $pixel_size 0 * $font * * * *]}] {
	puts sderr "Failed to get Text extents: font $font not found?"
	return
    }

    # determine how big our layer should be.
    set total_width 0
    for {set i 0} {$i < [string length $text]} {incr i} {
	 set total_width \
		 [expr \
		 [lindex [gimp-text-get-extents [string index $text $i] \
		 $pixel_size 0 * $font * * * *] 0] + \
		 $total_width + $spacing]
    }

    # if we're not being called from "build-ui" we create a new image the
    # appropriate size
    if ![info exists image] {
	.ctl.cq.progress configure -text "gimp-image-new"
	set img [gimp-image-new $total_width [lindex $extents 1] 0]
    } else {
	set img $image
    }
    # Create the layer, and add it to the image
    set main_layer [gimp-layer-new $img $total_width [lindex $extents 1] 0 \
	    "Spaced Out Text" 100 0]
    gimp-image-add-layer $img $main_layer 0

    # if the background should be transparent, make it so
    if {$transparent} {
	gimp-layer-add-alpha $main_layer
	gimp-drawable-fill $main_layer 2
    }

    # if the image is newly created, we fill in the background (if it's not
    # transparent), and create a new display
    if ![info exists image] {
	if !$transparent {
	    gimp-drawable-fill $main_layer 0
	}
	gimp-display-new $img
    }

    # our current x coord to write the text to
    set width 0

    # gimp-text-get-extents and gimp-text don't work with single spaces, so
    # we determine the width of a space by subtracting the width of "ab" from
    # the width of "a b".
    set space_width \
	    [expr [lindex [gimp-text-get-extents "a b" $pixel_size 0 * \
	    $font * * * *] 0] - \
	    [lindex [gimp-text-get-extents "ab" $pixel_size 0 * \
	    $font * * * *] 0]]
    # loop through each letter of the text, and print it.
    for {set i 0} {$i < [string length $text]} {incr i} {
	# if it's a space, we just add 'space_width' to 'width' and continue.
	if {[string trim [string index $text $i]] == ""} {
	    set width [expr $space_width + $width + $spacing]
	    continue
	}
	# get the extents for this letter.  Unfortunately, only the width is
	# different for each letter, not the height, ascent, or descent.
	set letter_extents [gimp-text-get-extents [string index $text $i] \
		$pixel_size 0 * $font * * * *]
	
	# create a new layer with the letter
	set layer [gimp-text $img $main_layer $width 0 "[string index $text $i]" 0 1 $pixel_size 0 * $font * * * *]

	# hokey: we will
	# use the "descent" value to determine where to place the letter on the
	# y axis.  if this is one of the "standard" droopy letters, we'll place
	# it lower according to the "descent" value of the font.
	if {[string first [string index $text $i] "jpqy"] != -1} {
	    gimp-layer-set-offsets $layer $width \
		    [expr [lindex $letter_extents 1] - \
		    [gimp-drawable-height $layer]]
	} else {
	    gimp-layer-set-offsets $layer $width \
		    [expr [lindex $letter_extents 1] - \
		    [gimp-drawable-height $layer] - [lindex $letter_extents 3]]
	}
	# anchor this new layer
	gimp-floating-sel-anchor $layer

	set width [expr [lindex $letter_extents 0] + $width + $spacing]
    }
    # flush the display
    gimp-displays-flush    
}	

gimp-main